//
//  MyAppController.m
//  LilyRunner
//
//  Created by Toshi Nagata on 09/02/07.
//  Copyright 2009 Toshi Nagata. All rights reserved.
//
/*
 Copyright (c) 2009 Toshi Nagata. All rights reserved.
 
 Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 
 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 
 THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#import "MyAppController.h"
#import "PreferenceWindowController.h"
#import "MyDocument.h"

#include <signal.h>
#include <unistd.h>
#include <sys/select.h>
#include <sys/types.h>
#include <sys/stat.h>

@implementation MyAppController

NSString *gLilyPondPathKey = @"lilypondpath";
NSString *gPreviewPathKey = @"previewpath";
NSString *gEditorPathKey = @"editorpath";
NSString *gEditCommandKey = @"editcommand";
NSString *gPNGResolutionKey = @"pngresolution";

- (id)defaultSettingForKey: (NSString *)key
{
	if ([key isEqualToString: gLilyPondPathKey]) {
		return @"/Applications/LilyPond.app";
	} else if ([key isEqualToString: gPreviewPathKey]) {
		return @"/Applications/Preview.app";
	} else if ([key isEqualToString: gEditorPathKey]) {
		return @"/Applications/TextEdit.app";
	} else if ([key isEqualToString: gEditCommandKey]) {
		return @"#!/bin/sh\n"
		"# $1 = file, $2 = line, $3 = column\n"
		"/usr/bin/osascript <<EOF\n"
		"tell application \"TextEdit\"\n"
		"  open (\"$1\" as POSIX file)\n"
		"  activate\n"
		"end tell\n"
		"EOF\n";
	} else if ([key isEqualToString: gPNGResolutionKey]) {
		return @"300";
	} else return @"";
}

//  Filter method for select Lilypond
- (BOOL)panel: (id)sender shouldShowFilename: (NSString *)filename
{
	NSFileManager *manager = [NSFileManager defaultManager];
	if ([[filename lastPathComponent] isEqualToString: @"app"]) {
		NSString *executable = [filename stringByAppendingString: @"/Contents/Resources/bin/lilypond"];
		return [manager isExecutableFileAtPath: executable];
	} else {
		return [filename rangeOfString: @"lilypond" options: NSCaseInsensitiveSearch].length > 0 &&
			[manager isExecutableFileAtPath: filename];
	}
}

- (BOOL)selectLily
{
	NSOpenPanel *panel;
	NSString *val;
	int retval;
	panel = [NSOpenPanel openPanel];
	[panel setCanChooseDirectories: YES];
	[panel setAllowsMultipleSelection: NO];
	[panel setMessage: @"Where is Lilypond?"];
	[panel setDelegate: self];
	retval = [panel runModal];
	[panel close];
	if (retval == NSFileHandlingPanelOKButton) {
		val = [panel filename];
		[[NSUserDefaults standardUserDefaults] setValue: val forKey: gLilyPondPathKey];
		return YES;
	}
	return NO;
}

- (BOOL)selectPreview
{
	NSOpenPanel *panel;
	NSString *val;
	int retval;
	panel = [NSOpenPanel openPanel];
	[panel setCanChooseDirectories: YES];
	[panel setAllowsMultipleSelection: NO];
	[panel setMessage: @"Which application to open pdf?"];
	retval = [panel runModalForTypes: [NSArray arrayWithObjects: @"app", nil]];
	[panel close];
	if (retval == NSFileHandlingPanelOKButton) {
		val = [panel filename];
		[[NSUserDefaults standardUserDefaults] setValue: val forKey: gPreviewPathKey];
		return YES;
	}
	return NO;
}

- (BOOL)selectEditor
{
	NSOpenPanel *panel;
	NSString *val;
	int retval;
	panel = [NSOpenPanel openPanel];
	[panel setCanChooseDirectories: YES];
	[panel setAllowsMultipleSelection: NO];
	[panel setMessage: @"Which application to edit lily scripts?"];
	retval = [panel runModalForTypes: [NSArray arrayWithObjects: @"app", nil]];
	[panel close];
	if (retval == NSFileHandlingPanelOKButton) {
		val = [panel filename];
		[[NSUserDefaults standardUserDefaults] setValue: val forKey: gEditorPathKey];
		return YES;
	}
	return NO;
}

- (void)enterPNGResolution: (id)sender
{
	int n = [sender intValue];
	if (n < 20 || n > 10000) {
		NSAlert *alert = [NSAlert alertWithMessageText:nil defaultButton:nil alternateButton:nil otherButton:nil informativeTextWithFormat:@"The PNG resolution should be 20..10000"];
		[alert runModal];
	} else {
		[[NSUserDefaults standardUserDefaults] setValue: [NSString stringWithFormat:@"%d", n] forKey: gPNGResolutionKey];
	}
}

- (BOOL)openEditorWithFile: (NSString *)path line: (int)line column: (int)column
{
	NSString *script, *cmd;
	const char *str;
	int fd, len;
	char tmpname[] = "/tmp/lilyRunnerScript_XXXX";
	
	if ((fd = mkstemp(tmpname)) < 0)
		goto error;
	script = [[NSUserDefaults standardUserDefaults] valueForKey: @"editcommand"];
	str = [script UTF8String];
	len = strlen(str) + 1;
	if (write(fd, str, len) < len)
		goto error;
	close(fd);
	if (chmod(tmpname, 0700) < 0)
		goto error;
	cmd = [NSString stringWithFormat: @"%s \"%s\" %d %d", tmpname, [path fileSystemRepresentation], line, column];
	NSLog(cmd);
	system([cmd UTF8String]);
	unlink(tmpname);
	return YES;
error:
	NSRunAlertPanel(@"Error", @"Cannot create temporary file", @"OK", nil, nil);
	return NO;
}

#pragma mark ==== Menu Commands ====

- (IBAction)openPreference: (id)sender
{
	id cont = [[PreferenceWindowController alloc] init];
	[[cont window] makeKeyAndOrderFront: self];
}

- (IBAction)openDocument: (id)sender
{
	//  Intercept the "Open..." command, which is usually handled by NSDocumentController
	NSDocumentController *dc = [NSDocumentController sharedDocumentController];
	NSArray *urls = [dc URLsFromRunningOpenPanel];
	NSEnumerator *en = [urls objectEnumerator];
	NSError *error;
	id obj, doc;
	while ((obj = [en nextObject]) != nil) {
		doc = [dc documentForURL: obj];
		if (doc == nil) {
			doc = [dc openUntitledDocumentAndDisplay: YES error: &error];
			if (doc == nil) {
				[[NSAlert alertWithError: error] runModal];
				return;
			}
			[doc setFileURL: obj];
		}
		[doc compile: self];
	/*
		if ([obj isFileURL] && (doc = [dc documentForURL: obj]) != nil) {
			[doc invokeLilypondWithPath: [obj path]];
		} else {
			[dc openDocumentWithContentsOfURL: obj display: YES error: NULL];
		}
	*/
	}
}

- (IBAction)openAndUpdateSyntax: (id)sender
{
	NSDocumentController *dc = [NSDocumentController sharedDocumentController];
	NSArray *urls = [dc URLsFromRunningOpenPanel];
	NSEnumerator *en = [urls objectEnumerator];
	NSError *error;
	id obj, doc;
	while ((obj = [en nextObject]) != nil) {
		doc = [dc documentForURL: obj];
		if (doc == nil) {
			doc = [dc openUntitledDocumentAndDisplay: YES error: &error];
			if (doc == nil) {
				[[NSAlert alertWithError: error] runModal];
				return;
			}
			[doc setFileURL: obj];
		}
		[doc updateSyntax: self];
	}
}

- (IBAction)openDocumentWithExternalEditor: (id)sender
{
	NSDocumentController *dc = [NSDocumentController sharedDocumentController];
	NSArray *urls = [dc URLsFromRunningOpenPanel];
	NSEnumerator *en = [urls objectEnumerator];
	NSError *error;
	id obj, doc;
	while ((obj = [en nextObject]) != nil) {
		doc = [dc documentForURL: obj];
		if (doc == nil) {
			doc = [dc openUntitledDocumentAndDisplay: YES error: &error];
			if (doc == nil) {
				[[NSAlert alertWithError: error] runModal];
				return;
			}
			[doc setFileURL: obj];
		}
		[self openEditorWithFile: [obj path] line: 1 column: 1];
	}
}

#pragma mark ==== Delegate Methods ====

- (void)applicationWillFinishLaunching:(NSNotification *)aNotification
{
	/*  Set this application default for "textedit" URL scheme  */
	LSSetDefaultHandlerForURLScheme(CFSTR("textedit"), CFSTR("net.homeip.alchemy.LilyRunner"));
	
	/*  Initialize the default settings  */
	{
		NSFileManager *manager = [NSFileManager defaultManager];
		NSUserDefaults *def = [NSUserDefaults standardUserDefaults];

		id val = [def valueForKey: gLilyPondPathKey];
		if (val == nil) {
			val = [self defaultSettingForKey: gLilyPondPathKey];
			[def setValue: val forKey: gLilyPondPathKey];
		}
		if (![manager isExecutableFileAtPath: [val stringByAppendingString: @"/Contents/Resources/bin/lilypond"]] && ![manager isExecutableFileAtPath: val])
			[self selectLily];

		val = [def valueForKey: gPreviewPathKey];
		if (val == nil) {
			val = [self defaultSettingForKey: gPreviewPathKey];
			[def setValue: val forKey: gPreviewPathKey];
		}
		if (![manager fileExistsAtPath: val])
			[self selectPreview];

		val = [def valueForKey: gEditorPathKey];
		if (val == nil) {
			val = [self defaultSettingForKey: gEditorPathKey];
			[def setValue: val forKey: gEditorPathKey];
		}
		if (![manager fileExistsAtPath: val])
			[self selectEditor];
		
		val = [def valueForKey: gEditCommandKey];
		if (val == nil) {
			val = [self defaultSettingForKey: gEditCommandKey];
			[def setValue: val forKey: gEditCommandKey];
		}

		val = [def valueForKey: gPNGResolutionKey];
		if (val == nil) {
			val = [self defaultSettingForKey: gPNGResolutionKey];
			[def setValue: val forKey: gPNGResolutionKey];
		}
		[def synchronize];
	}
	
}

- (void)applicationDidFinishLaunching:(NSNotification *)aNotification
{
	//  If no document is opened yet, then ask the user to select one
//	if ([[[NSDocumentController sharedDocumentController] documents] count] == 0)
//		[self openDocument: self];
}

- (BOOL)applicationShouldOpenUntitledFile: (NSApplication *)sender
{
	return NO;
}

- (BOOL)application:(NSApplication *)theApplication openFile:(NSString *)filename
{
	NSError *error;
	NSDocumentController *dc = [NSDocumentController sharedDocumentController];
	NSURL *fileURL = [NSURL fileURLWithPath: filename];
	id doc = [dc documentForURL: fileURL];
	if (doc == nil) {
		doc = [dc openUntitledDocumentAndDisplay: YES error: &error];
		if (doc == nil) {
			[[NSAlert alertWithError: error] runModal];
			return NO;
		}
		[doc setFileURL: fileURL];
	}
	[doc compile: self];
	return YES;
}

@end
